#include "Stdafx.h"
#include "Sys_cons.h"

//Private stuff
namespace
{
	enum
	{
		CMD_CVARLIST = 1,
		CMD_CMDLIST  = 2,
		CMD_TEST	 = 3
	};
	const int  CON_MAXARGSIZE  = 80;
	const int  CON_MAXCONFIGLINES = 256;
	const int  CON_MAXBUFFEREDCMDS = 32;
}

//======================================================================================
//======================================================================================

/*
======================================
Constructor
======================================
*/ 
CConsole::CConsole() : m_parms(CON_MAXARGSIZE)
{
	m_itCmd = 0;

	//open logfile
	m_hLogFile = ::CreateFile("debug.log",
							GENERIC_WRITE, 
							FILE_SHARE_READ, 0,
							CREATE_ALWAYS, 
							FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH, 0);

	RegisterCommand("cvarlist",CMD_CVARLIST, this);
	RegisterCommand("cmdlist", CMD_CMDLIST, this);
	RegisterCommand("cfunclist", CMD_CMDLIST,this);
	RegisterCommand("ctest", CMD_TEST, this);
}

/*
==========================================
Destructor
==========================================
*/
CConsole::~CConsole()
{
	//close log file
	if(m_hLogFile != INVALID_HANDLE_VALUE)
		::CloseHandle(m_hLogFile);
}

/*
===============================================
print a string to debugging window	
===============================================
*/
void CConsole::ComPrint(const char* text)
{
	if (!text)
		return;

	//write to log
	ulong bytesWritten =0;
	if(m_hLogFile != INVALID_HANDLE_VALUE)
		WriteFile(m_hLogFile,text, strlen(text), &bytesWritten, 0);

}

/*
=======================================
Try to execute a string in the console
=======================================
*/
bool CConsole::ExecString(const char *string)
{
	m_parms = string;
	const char * szfirstArg = m_parms.UnsafeStringTok(0);

	for(CmdList::iterator itcmd = m_lCmds.begin(); itcmd != m_lCmds.end(); itcmd ++)
	{
		if(strcmp(itcmd->name, szfirstArg)==0)
		{	itcmd->handler->HandleCommand(itcmd->id,m_parms);
			return true;
		}
	}

	for (CVarList::iterator it = m_lCVars.begin(); it != m_lCVars.end(); it++)
	{
		if(!strcmp((*it)->name,szfirstArg))
		{
			//only exec'ed IF function returns true
			if((*it)->flags & CVAR_READONLY)
				ComPrintf("%s is a read-only variable\n",  (*it)->name);
			else if((!(*it)->handler) || (*it)->handler->HandleCVar((*it),m_parms))
			{
				int numtokens = 0;
				if(numtokens = m_parms.NumTokens() > 1)
				{
					//char argVal[128];
					const char * argVal = m_parms.UnsafeStringTok(1);
					
					if((*it)->type != CVAR_STRING)
						(*it)->Set(argVal);
					else
					{
						char newstr[80];
						const char * arg=0;
						strcpy(newstr,argVal);
				
						for(int i=2,len=0;i<numtokens;i++)
						{
							arg = m_parms.UnsafeStringTok(i);
							len += strlen(arg+1);
							if(len >= 80)
								break;
							strcat(newstr," ");
							strcat(newstr,arg);
						}
						(*it)->Set(newstr);
					}

					if((*it)->flags & CVAR_LATCH)
						ComPrintf("%s will be changed on restart\n", (*it)->name);
					else
						ComPrintf("\"%s\" changed to \"%s\"\n",
									(*it)->name,(*it)->string);
					return true;
				}
			}
			ComPrintf("\"%s\" = \"%s\", default:\"%s\"\n",
						(*it)->name,(*it)->string, (*it)->default_string);
			return true;
		}
	}
	return false;
}

/*
==========================================
Register CVar
Initialize CVar and alphabetically add to list
==========================================
*/
void CConsole::RegisterCVar(CVarBase * var,	I_ConHandler * handler)
{
	if(handler)
		var->handler = handler;

	for(CVarList::iterator it = m_lCVars.begin(); it != m_lCVars.end(); it++)
	{
		if(strcmp((*it)->name, var->name) > 0)
			break;
	}

	//Find archived keyval and set it to that.
	CParms parms(CON_MAXARGSIZE);
	if(GetTokenParms(var->name, &parms))
	{
		if(parms.NumTokens() > 1)
			var->ForceSet(parms.UnsafeStringTok(1));
	}
	m_lCVars.insert(it,var);
}

/*
==========================================
Register CFunc
==========================================
*/
void CConsole::RegisterCommand(const char *cmdname,
							   HCMD id,
							   I_ConHandler * handler)
{
	for(CmdList::iterator it = m_lCmds.begin(); it != m_lCmds.end(); it++)
	{
		if(strcmp(it->name, cmdname) > 0)
			break;
	}

	//Find archived command in config, and exec that as well
	CParms parms(CON_MAXARGSIZE);
	if(GetTokenParms(cmdname, &parms))
		handler->HandleCommand(id,parms);
	
	m_lCmds.insert(it,CCommand(cmdname,id,handler));
}

//======================================================================================
//Console Commands
//======================================================================================
/*
==========================================
Handle Console command
==========================================
*/
void CConsole::HandleCommand(HCMD cmdId, const CParms &parms)
{
	switch(cmdId)
	{
	case CMD_CVARLIST:
		CVarlist(parms);
		break;
	case CMD_CMDLIST:
		CCmdList(parms);
		break;
	case CMD_TEST:
		CFunctest(parms);
		break;
	}
}

/*
======================================
Prints out all the CVars and their values
======================================
*/
void CConsole::CVarlist(const CParms &parms)
{
	ComPrintf("Console Variable Listing\n");
	ComPrintf("========================\n");

	CVarList::iterator it;
	
	if(parms.NumTokens() >=2)
	{
		const char * arg = parms.UnsafeStringTok(1);
		int len= strlen(arg);
		for(it = m_lCVars.begin(); it != m_lCVars.end(); it++)
		{
			if(strncmp((*it)->name,arg,len)==0)
				ComPrintf("\"%s\" is \"%s\"\n",(*it)->name,(*it)->string);
		}
	}
	else
	{	
		for(it = m_lCVars.begin(); it != m_lCVars.end(); it++)
			ComPrintf("\"%s\" is \"%s\"\n",(*it)->name,(*it)->string);
	}

}

/*
======================================
Prints out all the CVars and their values
======================================
*/
void CConsole::CCmdList(const CParms &parms)
{
	ComPrintf("Console Command Listing\n");
	ComPrintf("========================\n");

	CmdList::iterator it;

	if(parms.NumTokens() >=2)
	{
		const char * arg = parms.UnsafeStringTok(1);
		int len= strlen(arg);
		for(it = m_lCmds.begin(); it != m_lCmds.end(); it ++)
		{
			if(!strncmp(it->name,arg,len))
				ComPrintf("%s\n",it->name);
		}
	}
	else
	{
		for(it = m_lCmds.begin(); it != m_lCmds.end(); it ++)
			ComPrintf("%s\n",it->name);
	}
}

/*
=====================================
Util test func
=====================================
*/
void CConsole::CFunctest(const CParms &parms)
{	ComPrintf("test");
}



//======================================================================================
//Command line
//======================================================================================

void CConsole::AddCmdLineParm(const char * cmdLine)
{	m_cmdLineParms.push_back(std::string(cmdLine));
}

bool CConsole::IsCmdLineParm(const char * token, int tokenLen)
{
	for(StringList::iterator it = m_cmdLineParms.begin(); it != m_cmdLineParms.end(); it++)
	{
		if(strncmp(it->c_str(), token, tokenLen)==0)
			return true;
	}
	return false;
}

void CConsole::ExecCmdLine()
{
	for(StringList::iterator itVal = m_cmdLineParms.begin(); itVal != m_cmdLineParms.end(); itVal++)
	{
		if(!ExecString(itVal->c_str()))
			ComPrintf("Unknown command \"%s\"\n",itVal->c_str());
	}
}

//======================================================================================
//Config files etc
//======================================================================================

/*
======================================
Archived Token Access
======================================
*/
bool CConsole::GetTokenParms(const char * token, CParms * parms)
{
	int len = strlen(token);
	const char * archivedString=0;

	for(StringList::iterator itVal = m_configFileParms.begin(); itVal != m_configFileParms.end(); itVal++)
	{
		archivedString = itVal->c_str();
		if(strncmp(token, archivedString, len ) == 0)
		{
			parms->Set(itVal->c_str());
			return true;
		}
	}
	return false;
}

/*
======================================
Fill buffer with a parm from the config file
======================================
*/
int CConsole::ReadConfigParm(char * buf, int bufsize, FILE * fp)
{
	int len = 0;
	char c = fgetc(fp);

	if(c == EOF || c == '\0')
		return false;

	while(c && (c != '\n') && (c != EOF) && (len < CON_MAXARGSIZE))
	{
		buf[len++] = c;
		c = fgetc(fp);
	}
	buf[len] = '\0';
	return len;
}

/*
======================================
Load cvar data from a config file
======================================
*/
void CConsole::LoadConfig(const char * szFilename)
{
	FILE * fpcfg = fopen(szFilename,"r");
	if(fpcfg == NULL)
	{
		ComPrintf("CConsole::ExecConfig:Error opening %s\n",szFilename);
		return;
	}

	m_configFileParms.empty();

	int  lines=0;
	char line[CON_MAXARGSIZE];
	
	//Configs files are linited to 256 lines, 80 chars each
	do
	{
		if(!ReadConfigParm(line,CON_MAXARGSIZE,fpcfg))
			break;

		m_configFileParms.push_back(std::string(line));
		memset(line,0,CON_MAXARGSIZE);
		lines ++;

	}while(lines < CON_MAXCONFIGLINES);

	fclose(fpcfg);
	ComPrintf("CConsole::Loaded %s\n",szFilename);
}

/*
======================================
Runs a Config file
======================================
*/
void CConsole::ExecConfig(const char *szFilename)
{
	FILE * fpcfg=fopen(szFilename,"r");
	if(fpcfg == NULL)
	{
		ComPrintf("CConsole::ExecConfig:Error opening %s\n",szFilename);
		return;
	}

	//Configs files are linited to 256 lines, 80 chars each
	int  lines=0;
	const char * firstParm=0;
	int firstParmLen=0;
	char line[CON_MAXARGSIZE];
	CParms parm(80);
	
	memset(line,0,CON_MAXARGSIZE);
	
	do
	{
		if(!ReadConfigParm(line,CON_MAXARGSIZE,fpcfg))
			break;
		
		parm = line;
		firstParm = parm.UnsafeStringTok(0);
		firstParmLen = strlen(firstParm);

		if(IsCmdLineParm(firstParm,firstParmLen))
			continue;
		
		if(!ExecString(line))
			ComPrintf("Unknown command \"%s\"\n",line);
		memset(line,0,CON_MAXARGSIZE);
		lines ++;

	}while(lines < CON_MAXCONFIGLINES);

	fclose(fpcfg);
	ComPrintf("CConsole::Exec'ed %s\n",szFilename);
}

/*
=====================================
Write CVars to given file
=====================================
*/
void CConsole::WriteCVars(const char * szFilename)
{
	FILE * fp = fopen(szFilename,"w");
	if(!fp)
	{
		ComPrintf("CConsole::WriteCVars: Error writing config file\n");
		return;
	}

	//write all the archive flaged vars in the config file
	for (CVarList::iterator it = m_lCVars.begin(); it != m_lCVars.end(); it++)
	{
		if((*it)->flags & CVAR_ARCHIVE)
		{
			char line[80];
			(*it)->Unlatch();
			sprintf(line,"%s \"%s\"\n", (*it)->name, (*it)->string);
			fputs(line,fp);
		}
	}
	fclose(fp);
}

